/* klunar.c "wrapper" for Windows */

#include "precomp.h"

#include "platform.h"
#include "sapf.h"
#include "xrp.h"
#include "netbox.h"
#include "lunar.h"

static char *forwarding;
static int dgwmode;

struct sapfdest_s {
      int ifnr;
      struct sockaddr dst;
      sapfsel_t sel;
};

static int mysapfhandle;

static struct net_device *lunardev;
static struct net_device *basedev;

/* Note: In order to return a special device
   for the base device (other than the Lunar
   device, which has index 0) */
char basedevname[5] = {'B', 'A', 'S', 'E', '\0'};

static sapfsel_t ipv4deliver;
static sapfsel_t bcast_selector;

static int klunar_netdev_event(struct notifier_block *this,
			       unsigned long event, void *ptr);
static struct notifier_block klunar_notifier = {
  klunar_netdev_event, NULL, 0
};

// ----------------------------------------------------------------------

#include "knetbox.c"

static void L2sendto(sapfpkt *pkt, void *aux)
{
  struct sapfdest_s *dst = (struct sapfdest_s *) aux;

  if (!aux)
    return;

  DEBUG(10, printk("L2send %d bytes\n", pkt->len));

  netbox_l2send(dst->ifnr, &(dst->dst),
		dst->sel, pkt->data, pkt->len);
  return;
}

static void L3deliver(sapfpkt *pkt, void *aux)
{
  struct net_device_stats *stats;

  DEBUG(10, printk("L3deliver %d Bytes\n", pkt->len));

  stats = (struct net_device_stats *) lunardev->priv;
  if (stats) {
    stats->rx_packets++;
    stats->rx_bytes += pkt->len;
  }

  //	netbox_l3deliver(pkt->ifnr, pkt->data, pkt->len);
  netbox_l3deliver(lunardev->ifindex, pkt->data, pkt->len);

  return;
}

#include "../src/lunar.c"
#include "../src/lunar_dhcp.c"
#include "../src/xrp.c"

// ----------------------------------------------------------------------

static int klunar_open(struct net_device *dev)
{
  netif_start_queue(dev);
  return 0;
}


static int klunar_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
  struct net_device_stats *stats = (struct net_device_stats *) dev->priv;
  sapfsel_t sel;
  sapfpkt pkt;

  DEBUG(10, printk("klunar start_xmit (%d bytes)\n", skb->len));

  switch (ntohs(*(unsigned short *) (skb->data + 12))) {
  case ETH_P_ARP:
    DEBUG(10, printk("klunar - arp detected\n"));
    lunar_do_arp(*(struct in_addr *)(skb->data + 38));
    kfree_skb(skb);
    break;
  case ETH_P_IP:
    if (stats) {
      stats->tx_packets++;
      stats->tx_bytes += skb->len;
    }
    {
      int i;
      DEBUG(10, printk("IP data (%d):", skb->len));
      for (i = 0; i < 50; i++)
	DEBUG(10, printk(" %02x", skb->data[i]));
      DEBUG(10, printk("\n"));
    }
    pkt.data = skb->data + dev->hard_header_len;
    pkt.len = skb->len - dev->hard_header_len;
    pkt.ifnr = -1;
    
    if (pkt.len >= (20+8+44)  && *((char*)pkt.data+9) == 17 && // UDP
	ntohs(*(short*)((char*)pkt.data+20)) == 68 &&  // DHCP clnt
	ntohs(*(short*)((char*)pkt.data+22)) == 67)    // DHCP srvr
      lunar_do_dhcp(&pkt);
    else if (!memcmp(skb->data, "\xff\xff\xff\xff\xff\xff", ETH_ALEN)){ // broadcast
      if (bcast_selector)
	sapf_demux(mysapfhandle, bcast_selector, &pkt);
      else
	lunar_do_bc_setup(&bcast_selector, &pkt);
    } else { // unicast
      sel = 0;
      memcpy(((char*)&sel) + sizeof(sel) - ETH_ALEN,
	     skb->data,ETH_ALEN);
      sapf_demux(mysapfhandle, sel, &pkt);
    }
    kfree_skb(skb);
    break;
  default:
    kfree_skb(skb);
    break;
  }

  return 0;
}


static int klunar_stop(struct net_device *dev)
{
  if (basedev) {
    int i = basedev->ifindex;
    basedev = 0;
    netbox_dropL2if(i);
  }
  netif_stop_queue(dev);
  
  return 0;
}


static struct net_device_stats *klunar_get_stats(struct net_device *dev)
{
  return (struct net_device_stats *) dev->priv;
}


static int klunar_netdev_event(struct notifier_block *this,
			       unsigned long event, void *ptr)
{
  struct net_device *dev = (struct net_device *) ptr;

  if (!dev)
    return NOTIFY_DONE;

  DEBUG(10, printk("klunar notifier event %08lx %s\n", event, dev->name));

  switch (event) {
  case NETDEV_REGISTER:
  case NETDEV_UP:
    DEBUG(10, printk("notifier up %s\n", dev->name));
    if (basedev)
      break;
    if (basedevname) {
      struct net_device *d = dev_get_by_name(basedevname);
      if (!d)
	break;
      dev_put(d);
      if (d->ifindex != dev->ifindex)
	break;
      if (!strcmp(basedevname, dev->name)) {
	DEBUG(1, printk("klunar - new interface %s cannot be the lunar device\n", basedevname));
	break;
      }
    } else {
      if (!dev->get_wireless_stats)
	break;
    }
    DEBUG(1, printk("klunar - new (wireless) interface %s found\n",
		    dev->name));
    basedev = dev;
    netbox_addL2if(dev->ifindex, myhandle);
    break;
  case NETDEV_UNREGISTER:
  case NETDEV_DOWN:
    DEBUG(10, printk("notifier down %s\n", dev->name));
    if (dev == basedev) {
      DEBUG(1, printk("klunar - (wireless) interface %s went away\n",
		      dev->name));
      basedev = 0;
      netbox_dropL2if(dev->ifindex);
    }
    break;
    
  default:
    break;
  };

  return NOTIFY_DONE;
}

// ----------------------------------------------------------------------

int klunar_init(void)
{
  struct net_device *dev = 0;

  DEBUG(1, printk("klunar init module\n"));

  dgwmode = forwarding && !strcmp(forwarding, "yes");

  mysapfhandle = sapf_open("main", 641);
  if (mysapfhandle < 0) {
    DEBUG(10, printk("sapf_open failed\n"));
    return -1;
  }

  netbox_init();

  lunardev = init_etherdev(0, sizeof(struct net_device_stats));
  if (!lunardev) {
    DEBUG(10, printk("klunar_init: no lunardev\n"));
    sapf_close(mysapfhandle);
    return -ENOMEM;
  }
  SET_MODULE_OWNER(lunardev);
  DEBUG(1, printk("klunar - lunar device is %s\n", lunardev->name));

  if (basedevname) {
    dev = dev_get_by_name(basedevname);
    if (!dev) {
      DEBUG(1, printk("klunar - device %s not found\n", basedevname));
    } else {
      dev_put(dev);
      if (dev == lunardev) {
	DEBUG(1, printk("klunar - invalid base device %s\n",
			basedevname));
	dev = 0;
      }
    }
    basedev = dev;
    if (dev)
      netbox_addL2if(dev->ifindex, mysapfhandle);
  } else {
    read_lock(&dev_base_lock);
    for (dev = dev_base; dev != NULL; dev = dev->next) {
      if (dev->get_wireless_stats)
	break;
    }
    if (dev) {
      basedev = dev;
      DEBUG(10, printk("1, klunar - wireless interface is %s\n",
		       dev->name));
      netbox_addL2if(dev->ifindex, mysapfhandle);
    }
    read_unlock(&dev_base_lock);
  }
  
  sapf_addCallback(mysapfhandle, htonsel(SAPF_SELNET_LUNAR_RREQ),
		   lunar_rreq_handler, 0);
  sapf_setPermanent(mysapfhandle, htonsel(SAPF_SELNET_LUNAR_RREQ));
  
  ipv4deliver = sapf_getUnusedSelector(mysapfhandle, SELCLASS_LOCAL);
  sapf_addCallback(mysapfhandle, ipv4deliver, L3deliver, 0);
  sapf_setPermanent(mysapfhandle, ipv4deliver);
  
  lunardev->open = &klunar_open;
  lunardev->hard_start_xmit = &klunar_start_xmit;
  lunardev->stop = &klunar_stop;
  lunardev->get_stats = &klunar_get_stats;
  
  lunardev->mtu -= sizeof(sapfsel_t);
  lunardev->tx_queue_len = 3;
  *(short*)(lunardev->dev_addr) = htons(0x00ff);
  get_random_bytes(lunardev->dev_addr + 2, 4);
  
  register_netdevice_notifier(&klunar_notifier);
  proc_net_create("lunar", 0, lunar_proc_info);
  
  return 0;
}


void klunar_fini(void)
{
  DEBUG(1, printk("klunar unloading module\n"));

  unregister_netdevice_notifier(&klunar_notifier);
  unregister_netdev(lunardev);

  sapf_deleteEntry(mysapfhandle, htonsel(SAPF_SELNET_LUNAR_RREQ));
  sapf_deleteEntry(mysapfhandle, ipv4deliver);
  ipv4deliver = 0;

  if (basedev) {
    struct net_device *d = basedev;
    basedev = 0;
    netbox_dropL2if(d->ifindex);
  }

  netbox_fini();

  sapf_close(mysapfhandle);
}

// eof
