// knetbox.c

#include "precomp.h"
#include "sapf.h"
#include "netbox.h"

static struct packet_type sapfpacket;
static int myhandle = -1;
static char hid[NETBOX_HID_LEN];

// ----------------------------------------------------------------------

#define L2OUT_MS_PUMP_INTERVAL 1

/* Mutex counter used to synchronize message pump, in
   order to prevent new messages from being sent
   while we are sending other messages */
static LONG mutexCounter = 0;

static netbox_timer_t *timerlist;
static struct timer_list netbox_timer;

struct notifier_block *notifier_bl = NULL;

extern BYTE* GetMACAddress();
extern struct in_addr GetIPAddress();
extern PADAPT GetAdapter();

void interv_outg_ndis_traffic(BYTE* packet, int length);
int l2recv(struct sk_buff *skb, struct net_device *indev, struct packet_type *pt);

/* Spin lock used to protect queue of outgoing packets */
NDIS_SPIN_LOCK l2outLock;
/* List of pending packets to send */
LIST_ENTRY l2outQueue;

/* List of pending packets to call l2recv with */
NDIS_SPIN_LOCK l2rcvLock;
LIST_ENTRY l2rcvQueue;
/* List of pending packets to call interv_outg_ndis_traffic with */
NDIS_SPIN_LOCK l2intrvOutLock;
LIST_ENTRY l2intrvOutQueue;

NDIS_STATUS send_through_ndis(IN PADAPT pAdapt, PVOID packet)
{
  NDIS_STATUS status = NDIS_STATUS_SUCCESS;
  PNDIS_PACKET packetToSend = NULL;
  PNDIS_BUFFER packetToSendBuf = NULL;
  PQUEUED_PACKET qPacket = (PQUEUED_PACKET)packet;
  PSEND_RSVD rsvd;

  /* Allocate packet buffer descriptor (using the given
     packet buffer for which memory was already
     allocated) */
  NdisAllocateBuffer(&status, &packetToSendBuf, pAdapt->bufferPoolHandle,
		     qPacket->buffer, qPacket->length);

  /* If buffer storage could not be allocated,
     then quit the send */
  if (status != NDIS_STATUS_SUCCESS || packetToSendBuf == NULL)
    {
      DEBUG(2, printk("Netbox: Not enough pool space for NdisAllocateBuffer"));
      return NDIS_STATUS_FAILURE;
    }

  /* Allocate and initializes packet descriptor */
  NdisAllocatePacket(&status, &packetToSend, pAdapt->SendPacketPoolHandle);

  /* If everything went well */
  if (status == NDIS_STATUS_SUCCESS && packetToSend != NULL)
    {
      if (qPacket->direction == NDIS_DIR_SEND) {
	NdisSetPacketFlags(packetToSend, NDIS_FLAGS_DONT_LOOPBACK);
      }

      rsvd = (PSEND_RSVD)(packetToSend->ProtocolReserved);
      rsvd->OriginalPkt = 0;

      NdisChainBufferAtFront(packetToSend, packetToSendBuf);

      if (pAdapt != NULL && pAdapt->BindingHandle != NULL &&
	  packetToSend != NULL) {
	if (qPacket->direction == NDIS_DIR_RECEIVE) {
	  NDIS_SET_PACKET_STATUS(packetToSend, NDIS_STATUS_RESOURCES);
	  NdisMIndicateReceivePacket(pAdapt->MiniportHandle, &packetToSend, 1);
	} else if (qPacket->direction == NDIS_DIR_SEND) {
	  //DbgBreakPoint();
	  NdisSend(&status, pAdapt->BindingHandle, packetToSend);
	}

      } else {
	DEBUG(2, printk("Netbox: Adapter (or packet) was NULL!"));
      }

      /* Check how the send (or receive) went */
      switch (status)
	{
	  /* Everything went smoothly */
	case NDIS_STATUS_SUCCESS:
	  /* Free the allocated descriptors */
	  NdisFreeBuffer(packetToSendBuf);
	  NdisFreePacket(packetToSend);
	  break;
	case NDIS_STATUS_PENDING:
	  break;
	  /* TODO: Perhaps handle the error cases later */
	case NDIS_STATUS_INVALID_PACKET:
	case NDIS_STATUS_CLOSING:
	case NDIS_STATUS_FAILURE:
	default:
	  if (qPacket->direction == NDIS_DIR_RECEIVE) {
	    DEBUG(2, printk("Netbox: Ndis receive failure!"));
	  } else {
	    DEBUG(2, printk("Netbox: Ndis send failure!"));
	  }
	  break;
	}

    } else {
      /* If we get here that means we are out of
	 packets; Currently we then just drop the
	 packet */
      /* TODO: Could be dealt with by keeping separate
	 send and receive pools or (alternatively)
	 dynamically allocate more pools as needed and
	 free them when not needed */

      DEBUG(2, printk("Netbox: Failed to allocate Ndis packet!"));
    }

  /* Return the memory that had been allocated
     for the given packet buffer */
  kfree(qPacket->buffer);
  kfree(qPacket);

  return(status);
}

/* Callback function called on regular timer
   intervals to pump out messages in the L2
   out queue, i.e. messages going out to the
   network adapter */
VOID pump_L2out_queue(IN PVOID SystemSpecific1,
		      IN PVOID FunctionContext,
		      IN PVOID SystemSpecific2,
		      IN PVOID SystemSpecific3) {

  NDIS_STATUS status;
  PADAPT pAdapt = (PADAPT)FunctionContext;
  PLIST_ENTRY retrievedListEntry;
  PQUEUED_PACKET packetToSend, packetToLunar;

  if (NdisInterlockedIncrement(&mutexCounter) != 1) {
    /* Handler already active */
    DEBUG(2, printk("Netbox: Busy, cannot pump packets"));
    NdisInterlockedDecrement(&mutexCounter);
    return;
  }

  /* Check if there are any messages in the L2 out
     queue to pump out */
  /* Pump them all out */
  while ((retrievedListEntry = NdisInterlockedRemoveHeadList(&l2outQueue, &l2outLock)) != NULL) {
    DEBUG(10, printk("Netbox: Found message to pump in L2 out queue"));
    /* Retrieve pointer to the outer struct in
       which the list entry is contained */
    packetToSend = CONTAINING_RECORD(retrievedListEntry, QUEUED_PACKET, listEntry);

    /* Note: Allocated memory is freed when the packet
       has actually been sent */
    status = send_through_ndis(pAdapt, packetToSend);
  }


  /* Check if there are any messages in the l2recv
     queue to pump */
  /* Pump them all out */
  while ((retrievedListEntry = NdisInterlockedRemoveHeadList(&l2rcvQueue, &l2rcvLock)) != NULL) {
    packetToLunar = CONTAINING_RECORD(retrievedListEntry, QUEUED_PACKET, listEntry);
    /* TODO: Perhaps call the sapfpacket.func (to
       make everything more transparent) which
       should be set the l2recv function */
    l2recv((struct sk_buff*)packetToLunar->buffer, dev_get_by_index(0), &sapfpacket);
    /* Note: We can here release the memory used
       by the packet structure; The buffer has
       been freed in the receive function */
    kfree(packetToLunar);
  }

  /* Check if there are any messages in the l2intrvOut
     queue to pump */
  /* Pump them all out */
  while ((retrievedListEntry = NdisInterlockedRemoveHeadList(&l2intrvOutQueue, &l2intrvOutLock)) != NULL) {
    packetToLunar = CONTAINING_RECORD(retrievedListEntry, QUEUED_PACKET, listEntry);
    interv_outg_ndis_traffic((BYTE*)packetToLunar->buffer, packetToLunar->length);
    /* Note: We can here release the memory used
       by the packet structure; The buffer has
       been freed in the intervene function */
    kfree(packetToLunar);
  }

  NdisInterlockedDecrement(&mutexCounter);
}

/* NDIS worker thread used to pump packets */
VOID worker_thread (
    IN struct _NDIS_WORK_ITEM *  pWorkItem,
    IN PVOID  Context
    ) {
  pump_L2out_queue(NULL, Context, NULL, NULL);
  kfree(pWorkItem);
}

/* Use separate thread to handle packet pumping instead
   of timer interrupts since it is faster */
void activate_packet_pump(PADAPT pAdapt) {
  PNDIS_WORK_ITEM pWorkItem;

  pWorkItem = (PNDIS_WORK_ITEM)kmalloc(sizeof(NDIS_WORK_ITEM), 0);
  /* Note: If memory cannot be allocated for the worker
     thread, then the packet pump is not activated and
     the packet will stay on the queue until the next
     activation */
  if (pWorkItem) {
    NdisInitializeWorkItem(pWorkItem, (NDIS_PROC)worker_thread, (PVOID)pAdapt /* function context */);
    NdisScheduleWorkItem(pWorkItem);
  }
}

void set_notifier(struct notifier_block *nb)
{
  notifier_bl = nb;
}

void send_notification(unsigned long event)
{
  struct notifier_block *nb = notifier_bl;

  /* Never mind the priorities, we just notify
     everyone that is subscribing */
  while (nb != NULL) {
    (nb->notifier_call)(nb, event, dev_get_by_index(0));
    nb = nb->next;
  }
}

/* Callback function executed when the netbox timer
   has run out; Executes the next function in the queue
   and then starts the timer again */
static void timer_click(unsigned long dummy)
{
  if (timerlist) {
    netbox_timer_t *t = timerlist;
    timerlist = timerlist->next;
    if (timerlist) {
      netbox_timer.expires = timerlist->expires;
      add_timer(&netbox_timer);
    }
    t->fct(t->data);
  }
}

/* Creates a new timer (pointed to by t) which executes
   the funtion fct with the argument data after time
   delta_usec */
void netbox_newtimer(netbox_timer_t *t,
		     void (*fct)(void *), void *data, unsigned long delta_usec)
{
  netbox_timer_t **pp;
  unsigned long expires = jiffies + delta_usec / 1000; /*HZ * delta_usec / 1000000;*/

  DEBUG(10, printk("Netbox: newtimer %ld\n", delta_usec));
  if (!timerlist)
    pp = &timerlist;
  else {
    for (pp = &timerlist; ; pp = &((*pp)->next)) {
      if (!*pp || time_after((*pp)->expires, expires))
	break;
    }
  }
  t->next = *pp;
  t->expires = expires;
  t->fct = fct;
  t->data = data;
  *pp = t;

  if (t == timerlist) {
    if (t->next)
      del_timer(&netbox_timer);
    netbox_timer.expires = timerlist->expires;
    add_timer(&netbox_timer);
  }

  return;
}

/* Deletes the timer pointed to by t, i.e. cancels
   the executions of callback functions that may
   be pending */
void netbox_deltimer(netbox_timer_t *t)
{
  netbox_timer_t **pp;

  if (t == timerlist) {
    del_timer(&netbox_timer);
    timerlist = timerlist->next;
    if (timerlist) {
      netbox_timer.expires = timerlist->expires;
      add_timer(&netbox_timer);
    }
    return;
  }

  for (pp = &timerlist; *pp; pp = &((*pp)->next)) {
    if (*pp == t) {
      *pp = t->next;
      return;
    }
  }
}

// ----------------------------------------------------------------------

/* Function called by Ndis when it has received
   an outgoing IP or ARP packet (wrapped in an
   ethernet packet) */
void interv_outg_ndis_traffic(BYTE* packet, int length)
{
  static struct net_device *lunardev;
  struct sk_buff* skb = kmalloc(sizeof(struct sk_buff), 0);

  skb->nh.raw = packet;
  skb->len = length;
  skb->data = packet;

  lunardev = dev_get_by_index(0);

  if (lunardev) {
    lunardev->hard_start_xmit(skb, dev_get_by_index(0));
  }
}


/* Function used to send a packet wrapped in an
   ethernet packet; dest contains the ethernet
   address of the receiver, pkt_type contains the
   packet type (e.g. SAPF), data contains the
   packet data and len is the length of this data;
   Returns the amount of bytes from the given
   packet that were actually sent or -1 if the
   packet could not be sent */
int send_in_eth_pkt(struct sockaddr *dest, __u16 pkt_type, void *data, int len, int dir)
{
  PQUEUED_PACKET packetToSend = NULL;
  unsigned int eth_pkt_len;
  BYTE* eth_pkt;
  struct ether_header *eh;

  PADAPT pAdapt = GetAdapter();

  /* The given "data" is the packet which is to be
     wrapped in an ethernet packet; We create the
     (ethernet) packet to send and add it to the
     send queue */
  eth_pkt_len = len + sizeof(struct ether_header);

  eth_pkt = (BYTE*)kmalloc(eth_pkt_len, 0);
  if (!eth_pkt) {
    DEBUG(2, printk("Netbox: Could not allocate memory for send!"));
    return -1;
  }

  /* Queue the ethernet packet for later delivery */
  packetToSend = kmalloc(sizeof(QUEUED_PACKET), 0);
  if (!packetToSend) {
    kfree(eth_pkt);
    DEBUG(2, printk("Netbox: Could not allocate memory for send packet!"));
    return -1;
  }

  eh = (struct ether_header*) eth_pkt;
  /* Set ethernet destination address */
  memcpy(eh->ether_dhost, dest->sa_data, ETH_ALEN);
  /* Set ethernet packet type (to SAPF packet) */
  eh->ether_type = htons(pkt_type);
  /* Note: We ignore interface number; just send
     on the device we are interfacing, i.e. the
     source address of our ethernet packet is
     always the same one */
  memcpy(eh->ether_shost, GetMACAddress(), ETH_ALEN);

  /* Copy the packet contents to the first position
     after the ethernet header */
  memcpy(eth_pkt + sizeof(struct ether_header), (BYTE*)data, len);

  packetToSend->buffer = eth_pkt;
  packetToSend->length = eth_pkt_len;
  packetToSend->direction = dir;

  NdisInitializeListHead(&packetToSend->listEntry);
  NdisInterlockedInsertTailList(&l2outQueue, &packetToSend->listEntry, &l2outLock);
  /* Indicate that a handler thread should take
     the packet */
  activate_packet_pump(pAdapt);

  return len;
}


int netbox_init()
{
  PADAPT pAdapt = GetAdapter();

  if (!pAdapt) return -1;

  /* Start timer that will execute callback function */
  init_timer(&netbox_timer);
  netbox_timer.function = timer_click;

  get_random_bytes(hid, sizeof(hid));

  NdisAllocateSpinLock(&l2outLock);
  NdisAllocateSpinLock(&l2rcvLock);
  NdisAllocateSpinLock(&l2intrvOutLock);

  NdisInitializeListHead(&l2outQueue);
  NdisInitializeListHead(&l2rcvQueue);
  NdisInitializeListHead(&l2intrvOutQueue);

  return 0;
}


/* Callback function called when SAPF L2 traffic
   comes in, i.e. (ethernet) traffic from the
   network card with packet type SAPF set. This
   function is the same in the Linux and Windows
   versions. */
int l2recv(struct sk_buff *skb, struct net_device *indev, struct packet_type *pt)
{
  struct sapfpkt_s *p;
  sapfsel_t dest;

  p = (struct sapfpkt_s *) &skb->cb;
  memset(p, 0, sizeof(*p));
  p->snr = 0;
  p->ifnr = (short)indev->ifindex;
  p->data = skb->nh.raw + sizeof(sapfsel_t);
  p->len = skb->len - sizeof(sapfsel_t);

  memcpy(&dest, skb->nh.raw, sizeof(dest));

  DEBUG(10, printk("Netbox: sapf packet, %d (%d) bytes via %s: sel=0x%016Lx, refcnt=%d\n",
		   skb->len, skb->len - sizeof(sapfsel_t),
		   indev ? indev->name : "no-dev", ntohsel(dest),
		   indev->refcnt.counter));

  if (myhandle == -1) {
    DEBUG(1, printk("Netbox: Handle was not set, could not demux sel"));
  } else {
    sapf_demux(myhandle, dest, p);
    /* sapf_demux(*(int*)pt->data, dest, p); */
  }

  kfree_skb(skb);
  return 0;
}


/* Function called when an L2 network interface
   (with a given number) is added (i.e. comes up);
   In the Linux version a callback function handling
   SAPF L2 traffic, i.e. traffic coming in from the
   network adapter (ethernet packet with packet type
   SAPF), is added; In the Windows version of the
   code this function is (almost) the same, but we
   use the information differently while making sure
   that l2recv is called when SAPF packets come in. */
int netbox_addL2if(int ifnr, int handle)
{
  /*
  if (sapfpacket.func)
    return -1;
  */

  DEBUG(10, printk("Netbox: netbox addL2if, handle: %d\n", handle));
  
  myhandle = handle;

  sapfpacket.type = htons(SAPF_ETHERTYPE);
  sapfpacket.dev = dev_get_by_index(ifnr);
  sapfpacket.func = l2recv;
  sapfpacket.data = &myhandle;
  /* dev_add_pack(&sapfpacket); */
  
  return 0;
}


/* Function called when an L2 network interface
   (with a given number) is removed; In the Linux
   version the callback function handling SAPF L2
   traffic is dropped; In the Windows version
   however, we don't do anything. */
int netbox_dropL2if(int ifnr)
{
  DEBUG(10, printk("Netbox: netbox dropL2if\n"));

  /* TODO: Perhaps include this, used to report
     that we have stopped handling the L2 traffic
     through Lunar */
  /* neigh_ifdown(&arp_tbl, sapfpacket.dev); */

  return 0;
}

int netbox_getHID128(unsigned char *buf)
{
  memcpy(buf, hid, sizeof(hid));
  return 0; // success
}


/* Function used to retrieve our own network (i.e.
   ethernet) address; In the Linux version the
   network address of a specified network interface
   is returned but in the Windows version we get
   the nw address of the wireless network card for
   which we have installed this intermediate
   driver. */
int netbox_getL2addr(int ifnr, struct sockaddr *sa)
{
  int i;

  memcpy(sa->sa_data, GetMACAddress(), ETH_ALEN);

  DEBUG(10, printk("Netbox: getL2addr:"));
  for (i = 0; i < ETH_ALEN; i++)
    DEBUG(10, printk("%02x ", (unsigned char) sa->sa_data[i]));
  DEBUG(10, printk("\n"));

  return 0;
}


/* Function used to retrieve our own layer 3
   (i.e. IP layer) address; In the Linux version
   the first IP address pertaining to a certain
   network device is returned but in the Windows
   version we get the IP address pertaining to
   the wireless network card for which we have
   installed this intermediate driver. */
int netbox_getL3addr(struct in_addr *name, struct net_device *dev)
{

  name->s_addr = GetIPAddress().s_addr;
  DEBUG(10, printk("Netbox: local %08x\n", ntohl(name->s_addr)));

  return 0;
}


/* Function used to send an ARP announce
   (an ARP reply without an ARP request) */
int netbox_arpannounce(struct in_addr addr, struct net_device *dev)
{
  struct arp_pkt arp_packet;
  struct sockaddr broadcast_hw_addr;
  struct sockaddr dest;
  struct in_addr our_ip;
  int ignore;
	
  DEBUG(10, printk("Netbox: arpannounce: %08x\n", ntohl(addr.s_addr)));

  // TODO: We don't do anything here.. Not necessary?
  return 0;
}


/* Function called in order to update the local
   system ARP table with a new entry */
// IP addr in network order
int netbox_arpset(struct in_addr addr, struct sockaddr *hw_addr, struct net_device *dev)
{
  int i;
  struct arp_pkt arp_packet;
  struct sockaddr dest;
  struct in_addr our_ip;
  int ignore;

  DEBUG(10, printk("Netbox: arpset: %08x ", ntohl(addr.s_addr)));
  for (i = 0; i < ETH_ALEN; i++) {
    DEBUG(10, printk(" %02x", (unsigned char) hw_addr->sa_data[i]));
  }
  DEBUG(10, printk("\n"));

  /* Store ARP table entry in local ARP table
     by generating ARP reply destined for
     ourself */
  memcpy(dest.sa_data, GetMACAddress(), ETH_ALEN);

  arp_packet.ar_hrd = htons(0x0001);
  arp_packet.ar_pro = htons(ETH_P_IP);
  arp_packet.ar_hln = 0x06; /* Ethernet address size (in bytes) */
  arp_packet.ar_pln = 0x04; /* IP address size (in bytes) */
  arp_packet.ar_op = htons(ARPOP_REPLY);

  our_ip = GetIPAddress();
  memcpy(arp_packet.ar_sha, hw_addr->sa_data, ETH_ALEN);
  memcpy(arp_packet.ar_sip, &addr, sizeof(__u32));
  memcpy(arp_packet.ar_tha, GetMACAddress(), ETH_ALEN);
  memcpy(arp_packet.ar_tip, &(our_ip.s_addr), sizeof(__u32));

  /* Create and send ethernet packet wrapping
     the ARP packet */
  ignore = send_in_eth_pkt(&dest, ETH_P_ARP, &arp_packet,
			   sizeof(struct arp_pkt), NDIS_DIR_RECEIVE);

  return 0;
}


/* Function called in order to delete an entry from
   the local system ARP table */
// IP addr in network order
int netbox_arpdelete(struct in_addr addr, struct net_device *dev)
{
  struct sockaddr hw_addr;

  DEBUG(10, printk("Netbox: arpdelete %08x\n", ntohl(addr.s_addr)));

  /* TODO: Currently we rely on the Windows ArpCacheLife
     parameter in order to delete ARP table entries */
  return 0;
}


/* Send data out to the network card */
int netbox_l2send(int ifnr, struct sockaddr *dest, sapfsel_t sel,
		  void *data, int len)
     // selector must be in network order
{
  BYTE* sapf_pkt;
  unsigned int sapf_pkt_len = 0;
  int retval;

  static struct sockaddr broadcast = {AF_UNSPEC, {0xff, 0xff, 0xff, 0xff, 0xff, 0xff}};

  if (!dest)
    dest = &broadcast;

  DEBUG(10,
	printk("netbox_l2send %02x:%02x:%02x:%02x:%02x:%02x 0x%016Lx\n",
	       (unsigned char) dest->sa_data[0],
	       (unsigned char) dest->sa_data[1],
	       (unsigned char) dest->sa_data[2],
	       (unsigned char) dest->sa_data[3],
	       (unsigned char) dest->sa_data[4],
	       (unsigned char) dest->sa_data[5], 
	       ntohsel(sel)));

  sapf_pkt_len = len + sizeof(sapfsel_t);
  sapf_pkt = (BYTE*)kmalloc(sapf_pkt_len, 0);

  if (!sapf_pkt) {
    DEBUG(2, printk("Netbox: Could not allocate memory for SAPF packet!"));
    return -1;
  }

  /* Copy SAPF selector to the first position
     of the packet */
  memcpy(sapf_pkt, &sel, sizeof(sapfsel_t));
  /* Copy the packet contents to after the SAPF
     selector */
  memcpy(sapf_pkt + sizeof(sapfsel_t), (BYTE*)data, len);

  /* Create and send ethernet packet wrapping
     the SAPF packet */
  retval = send_in_eth_pkt(dest, SAPF_ETHERTYPE, sapf_pkt,
			   sapf_pkt_len, NDIS_DIR_SEND);

  kfree(sapf_pkt);
  return retval;
}


/* Send data further up the stack (to the IP layer),
   as if it came from the network card */
int netbox_l3deliver(int ifnr, void *data, int len)
{
  struct ether_header *eh = NULL;
  struct sockaddr dest;
  int ignore;

  DEBUG(10, printk("netbox_l3deliver %d Bytes\n", len));

  memcpy(dest.sa_data, GetMACAddress(), ETH_ALEN);

  /* Create and send ethernet packet wrapping
     the IP packet up the stack */
  ignore = send_in_eth_pkt(&dest, ETH_P_IP, data,
			   len, NDIS_DIR_RECEIVE);

  return 0;
}

/* int netbox_hasDGW(void);

   returns true when the host has a default GW entry in its
   routing table and should forward packets to the fixed
   internet.
*/
int netbox_hasDGW(void)
{
  return 0;
}


// ----------------------------------------------------------------------

void netbox_fini()
{
  if (timerlist)
    del_timer(&netbox_timer);

  NdisFreeSpinLock(&l2outLock);
  NdisFreeSpinLock(&l2rcvLock);
  NdisFreeSpinLock(&l2intrvOutLock);
}

// eof
